open Data
(* Box office analysis *)

(* Contents:
    -- the movie type
    -- the studio_gross type
    -- functions for querying and transforming lists of movies
*)

(* a movie is a tuple of (title, studio, gross in millions, year) *)
type movie = string * string * float * int

(* a studio_gross is a pair of (studio, gross in millions) *)
type studio_gross = string * float

(* call bad_argument if your function receives a bad argument *)
(* do not change this exception or function                   *)
exception Bad_arg of string
let bad_arg (s:string) = raise (Bad_arg s)

(* a useful debugging routine *)
let debug s = print_string s; flush_all()



(* Testing functions. You do not need to change these,
 * however, you are expected to use them to test your
 * functions. *)

(* Dummy test functions. No Need to edit these. *)
let average_tests () = ()
let decade_tests () = ()
let take_tests () = ()
let drop_tests () = ()
let selection_sort_tests () = ()
let sort_tests () = ()
let studio_tests () = ()


(* Function to compare floats which accounts for rounding error.*)
let cmp_float (f1:float) (f2:float) : bool =
  let d = f1 -. f2 in
  abs_float d < 0.0000000001

(* Compares the name of two movies *)
let compare_movie ((a, _, _, _) : movie) ((b, _, _, _) : movie) =
  compare a b

(* Compared the name of two studios*)
let compare_studio ((a, _) : studio_gross) ((b, _) : studio_gross) =
  compare a b

(* Prints a list of integers*)
let print_list_int (l : int list) =
  print_string "[";
  let rec iter l =
    match l with
      | a::b -> print_string "; "; print_int a; iter b
      | _ -> print_string "]"
  in
  match l with
    | a::b -> print_int a; iter b
    | _ -> print_string "]"

(* Prints a movie *)
let print_movie ((title, studio, gross, year) : movie) =
  print_string "(";
  print_string title;
  print_string ", ";
  print_string studio;
  print_string ", ";
  print_float gross;
  print_string ", ";
  print_int year;
  print_string ")"

(* Prints a list of movies *)
let print_data (l : movie list) =
  print_string "[";
  let rec iter l =
    match l with
      | a::b -> print_string "; "; print_movie a; iter b
      | _ -> print_string "]"
  in
  match l with
    | a::b -> print_movie a; iter b
    | _ -> print_string "]"

(* Prints a list of studios *)
let print_studio (l : studio_gross list) =
  print_string "[";
  let rec iter l =
    match l with
      | (x, y)::b -> print_string "; "; print_string x; iter b
      | _ -> print_string "]"
  in
  match l with
    | (x, y)::b -> print_string x; iter b
    | _ -> print_string "]"

(* Function to run tests. The arguments are a comparison function,
 * then your code's output, then the expected output, then a function
 * to print output, then finally the name of the test. Take note that
 * this code does not check for exceptions, so if your code throws an
 *  exception it will fail at runtime. *)
let run_test (cmp_func: 'a->'a->bool) (student_ans:'a) (correct_ans:'a)
    (print_fun:'a->unit) (test_name:string) =
  if (cmp_func student_ans correct_ans) == false then
    (print_string ("TEST FAILED: "^test_name^"\n");
    print_string "Got: ";
    print_fun student_ans;
    print_string "\nExpected: ";
    print_fun correct_ans;
    print_string "\n";
    ())
  else ()

(* Function for tests that are designed to fail. The arguments are a
 * student function that takes two arguments, the first argument, the
 * the second argument, a print function, and finally the name of the
 * test. These tests will fail if the function given does NOT throw a
 * Bad Arg exception for the arguments given. *)
let test_fail (student_fun: 'a -> 'b -> 'c) (arg : 'a) (arg2 : 'b)
  (print_fun:'c->unit) (test_name:string) =
  try
    let ans = student_fun arg arg2 in
    (print_string ("TEST FAILED: "^test_name^"\n");
    print_string "Got: ";
    print_fun ans;
    print_string "\nExpected: Bad Arg\n";
    ())
  with
  Bad_arg s  -> ()
  | _ -> (print_string ("TEST FAILED: "^test_name^"\n");
    print_string "Test does not throw correct exception\n")

(* *** DO NOT CHANGE DEFINITIONS ABOVE THIS LINE! *** *)

(* you may add "rec" after any of the let declarations below that you
 * wish if you find doing so useful. *)


(* find the average gross of the movies in the list                  *)
(* return 0.0 if the list is empty                                   *)
(* hint: you may need to use functions float_of_int and int_of_float *)
(* hint: if you don't know what those functions do,                  *)
(*       type them in to ocaml toplevel                              *)
(* hint: recall the difference between +. and + also 0. and 0        *)
let average (movies : movie list) : float = 
  failwith "average unimplemented"

(* Average tests. You should add more to these tests, as
 * the given tests are incomplete. Recall that the run_test
 * function takes a comparison function, a student answer,
 * the expected answer, a print method, and a test name. For
 * tests and the remaining problems in query.ml, you will
 * find it useful to use the examples defined in data.ml.
 * Feel free to add more examples to data.ml if you feel that
 * the given examples are insufficient.*)
(*
let average_tests () =
  run_test cmp_float (average []) 0.0 print_float "[]";
  run_test cmp_float (average data1) 377.85 print_float "data1";
  ()
*)

(* return a list containing only the movies from the given decade *)
(* call bad_arg if n is not 20, 30, ..., 90, 00, 10               *)
(* Treat 0 as 00 (this is unavoidable as 00 is not represented    *)
(*   differently from 0).                                         *)
(* Note: movies from any years outside the range 1920-2019 will   *)
(* always be discarded but should not raise an error condition    *)

let decade (n:int) (ms:movie list) : movie list =
  failwith "decade unimplemented"

(* Decade tests. You should add more to these tests, as
 * the given tests are incomplete. For this problem you may
 * also find the test_fail method useful. This method
 * requiers a student function, two arguments, and
 * a test name, and expects the function to throw a BAD ARG
 * Exception for the given arguments.
 * Take note of the operator (=). The parenthesis change the
 * usage of this operator from a = b to (=) a b. Another thing
 * to note is that as decade does not have to return arguments
 * in a specific order, the results are sorted to ensure consistency.*)
(*
let decade_tests () =
  run_test (=) (List.sort compare_movie (decade 00 data3))
    (List.sort compare_movie decade00from3) print_data "00's from data3";
  test_fail decade (-10) (data1) print_data "test negitive year";
  ()
*)

(* return the first n items from the list *)
(* if there are fewer than n items, return all of them *)
(* call bad_arg if n is negative *)
let rec take (n:int) (l:'a list)  : 'a list =
  failwith "take unimplemented"

(* Take tests. You should add more to these tests, as
 * the given tests are incomplete.*)
(*
let take_tests () =
  run_test (=) (take 1 data1) data1 print_data "1 from data1";
  ()
*)

(* return everything but the first n items from the list *)
(* if there are fewer than n items, return the empty list *)
(* call bad_arg if n is negative *)
let rec drop (n:int) (l:'a list)  : 'a list =
  failwith "drop unimplemented"

 (* Drop tests. You should add more to these tests, as
  * the given tests are incomplete.*)
(*
let drop_tests () =
  run_test (=) (drop 1 data4) drop1from4 print_data "1 from data4";
  ()
*)

(* return a list [x1; x2; ...; xn] with the same elements as the input l
   and where:
     leq xn xn-1
     ...
     leq x3 x2
     leq x2 x1
     are all true
*)
(* hint: define an auxiliary function "select" *)
type 'a less = 'a -> 'a -> bool
let selection_sort (leq:'a less) (l:'a list) : 'a list =
  failwith "selection_sort unimplemented"


(* Selection sort tests. You should add more to these tests, as
 * the given tests are incomplete.*)
 (*
let selection_sort_tests () =
  run_test (=) (selection_sort (<=) [1;2]) [2;1] print_list_int "[1;2]";
  ()
*)

(* ASIDE:  Why does this assignment ask you to implement selection sort?
   Insertion sort is almost always preferable to selection sort,
   if you have to implement a quadratic-time sorting algorithm.
   Insertion sort is faster, it's simpler to implement, and it's
   easier to reason about.  For smallish inputs (less than 5 or 8),
   insertion sort is typically faster than quicksort or any
   other NlogN sorting algorithm.  So, why do we ask you to implement
   selection sort?  Answer: we already showed you insertion sort
   in the lecture notes.

   ASIDE 2: But at least selection sort is better than bubble sort.
   Even Barack Obama knows that. https://www.youtube.com/watch?v=k4RRi_ntQc8
*)

(* return list of movies sorted by gross (largest gross first) *)
let sort_by_gross (movies : movie list) : movie list = 
  failwith "sort_by_gross unimplemented"


(* return list of movies sorted by year produced (largest year first) *)
let sort_by_year (movies : movie list) : movie list = 
  failwith "sort_by_year unimplemented"


(* sort list of (studio, gross in millions) by gross in millions 
 * with the largest gross first *)
let sort_by_studio (studio_grosses : studio_gross list) : studio_gross list = 
  failwith "sort_by_studio unimplemented"


(* given list of movies,
 * return list of pairs (studio_name, total gross revenue for that studio)  *)
let by_studio (movies:movie list) : studio_gross list =
  failwith "by_studio unimplemented"

 (* Tests for all sort functions.You should add more to these tests, as
  * the given tests are incomplete.*)
(*
let sort_tests () =
  run_test (=) (sort_by_gross data1) data1 print_data "data1 by gross";
  run_test (=) (sort_by_year data1) data1 print_data "data1 by year";
  run_test (=) (sort_by_studio [s2;s3;s1]) sOrdered
    print_studio  "[s2;s3;s1] by studio";
  ()
*)

(* given list of movies,
 * return list of pairs (studio_name, total gross revenue for that studio)  *)
let by_studio (movies:movie list) : studio_gross list =
  failwith "by_studio unimplemented"

 (* By_studio tests. You should add more to these tests, as
  * the given tests are incomplete.*)
(*
let studio_tests () =
  run_test (=) (List.sort compare_studio (by_studio data2))
    bs2 print_studio "data2 by studio";
  ()
*)
